/*----------------------------------------------------------------------
  File    : mpx.c
  Contents: multivariate polynomial regression execution
  Author  : Christian Borgelt
  History : 2007.04.12 file created from file mlpx.c
            2007.05.18 execution time measurements added
            2011.07.27 adapted to modified attset and utility modules
            2013.06.08 missing flag AS_MARKED added to table read mode
            2013.08.12 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.13 adapted to definition of type DIMID in matrix.h
            2013.08.20 output formats changed to significant digits
            2015.08.03 adapted to changed regression object interface
            2018.09.13 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef REG_PARSE
#define REG_PARSE
#endif
#include "regress.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "mpx"
#define DESCRIPTION "multivariate polynomial regression execution"
#define VERSION     "version 3.0 (2018.09.13)         " \
                    "(c) 2007-2018   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_ATTCNT    (-10)       /* no usable attributes found */
#define E_VECSIZE   (-11)       /* invalid pattern size */
#define E_OUTPUT    (-12)       /* target in input or write output */
#define E_TARGET    (-13)       /* invalid target attribute */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction result --- */
  ATT   *att;                   /* target attribute */
  int    type;                  /* type of the target attribute */
  INST   pred;                  /* predicted value */
  CCHAR *col_pred;              /* name   of prediction column */
  int    cwd_pred;              /* width  of prediction column */
  int    dig_pred;              /* digits of prediction column */
  double conf;                  /* confidence of predicted value */
  CCHAR *col_conf;              /* name   of confidence column */
  int    cwd_conf;              /* width  of confidence column */
  int    dig_conf;              /* digits of confidence column */
  double err;                   /* error value (squared difference) */
} RESULT;                       /* (prediction result) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_ATTCNT  -10 */  "no (usable) attributes (need at least 1)",
  /* E_VECSIZE -11 */  "invalid pattern size %"DIMID_FMT,
  /* E_OUTPUT  -12 */  "must have target as input or write output",
  /* E_TARGET  -13 */  "invalid target attribute %s",
  /*           -14 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static ATTMAP   *attmap = NULL; /* attribute map */
static TABLE    *table  = NULL; /* data table */
static REGRESS  *reg    = NULL; /* regression object */
static RESULT   res     = {     /* prediction result */
  NULL, AT_NOM,                 /* target attribute */
  {0}, "reg", 0, 6,             /* data for prediction column */
  0.0, NULL,  0, 6,             /* data for confidence column */
  0 };                          /* error value (squared difference) */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (reg)    reg_delete(reg);       \
  if (table)  tab_delete(table,  0); \
  if (attmap) am_delete(attmap,  0); \
  if (attset) as_delete(attset);     \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1); \
  if (scan)   scn_delete(scan,   1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void predict (REGRESS *reg, double *vec, double thresh)
{                               /* --- classify the current tuple */
  INST   *inst;                 /* to access the target instance */
  double pred;                  /* prediction of regression model */

  assert(reg);                  /* check for a regression model */
  pred = reg_exec(reg, vec);    /* execute regression polynomial */
  inst = att_inst(res.att);     /* and get the target instance */
  if      (res.type == AT_NOM){ /* if the target att. is nominal */
    if (pred >= thresh) { res.pred.n = 1; res.conf = 1.0 -pred; }
    else                { res.pred.n = 0; res.conf =      pred; }
    if      (res.conf > 1.0) res.conf = 1.0;
    else if (res.conf < 0.0) res.conf = 0.0;
    res.err = ((inst->n > NV_NOM) && (res.pred.n != inst->n)) ? 1 : 0; }
  else if (res.type == AT_INT){ /* if the target att. is integer */
    res.pred.i = (DTINT)(pred +0.5);
    res.err    = (!isnull(inst->i)) ? (double)(res.pred.i -inst->i) : 0;
    res.err   *= res.err; }     /* squared difference to true value */
  else {                        /* if the target att. is real-valued */
    res.pred.f = (DTFLT)pred;   /* convert the predicted value */
    res.err    = (!isnan(inst->f))  ? (double)(pred       -inst->f) : 0;
    res.err   *= res.err;       /* squared difference to true value */
  }
}  /* predict() */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, TABWRITE *twrite, int mode)
{                               /* --- write additional information */
  int n, k;                     /* character counters */

  assert(set && twrite);        /* check the function arguments */
  if (mode & AS_ATT) {          /* if to write the header */
    twr_puts(twrite, res.col_pred); /* write prediction column name */
    if ((mode & AS_ALIGN)       /* if to align the column */
    && ((mode & AS_WEIGHT) || res.col_conf)) {
      n = (int)strlen(res.col_pred);
      k = att_valwd(res.att,0); /* compute width of prediction column */
      res.cwd_pred = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      if (k > n) twr_pad(twrite, (size_t)(k-n));
    }                           /* pad with blanks if requested */
    if (res.col_conf) {         /* if to write a confidence field */
      twr_fldsep(twrite);       /* write a field separator and */
      twr_puts(twrite, res.col_conf);       /* the column name */
      if ((mode & AS_ALIGN) && (mode & AS_WEIGHT)) {
        n = (int)strlen(res.col_conf);
        k = res.dig_conf +3;    /* compute width of confidence column */
        res.cwd_conf = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* pad with blanks if requested */
    } }
  else {                        /* if to write a normal record */
    n = (res.type == AT_NOM)    /* get the class value */
      ? twr_printf(twrite, "%s", att_valname(res.att, res.pred.n))
      : (res.type == AT_INT)    /* format numeric types */
      ? twr_printf(twrite, "%"DTINT_FMT,         res.pred.i)
      : twr_printf(twrite, "%.*g", res.dig_pred, res.pred.f);
    if (res.cwd_pred > n) twr_pad(twrite, (size_t)(res.cwd_pred-n));
    if (res.col_conf) {         /* if to write a confidence field */
      twr_fldsep(twrite);       /* write separator and confidence */
      n = twr_printf(twrite, "%.*g", res.dig_conf, res.conf);
      if (res.cwd_conf > n) twr_pad(twrite, (size_t)(res.cwd_conf-n));
    }                           /* if to align, pad with blanks */
  }
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_reg  = NULL;      /* name of regression model file */
  CCHAR   *fn_tab  = NULL;      /* name of input  pattern file */
  CCHAR   *fn_out  = NULL;      /* name of output pattern file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  int     matinp   = 0;         /* flag for numerical matrix input */
  double  thresh   = 0.5;       /* classification threshold */
  DIMID   dim      = -1;        /* data point/pattern dimension */
  int     mode     = AS_ATT|AS_MARKED; /* table file read  mode */
  int     mout     = AS_ATT;           /* table file write mode */
  double  sse      = 0.0;       /* (weighted) sum of squared errors */
  double  *vec;                 /* to traverse the patterns */
  ATTID   a, c;                 /* number of attributes */
  DIMID   m;                    /* number of dimensions/fields */
  DIMID   n;                    /* number of data tuples */
  double  w, u;                 /* weight of data tuples, buffer */
  clock_t t;                    /* for time measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] regfile [-d|-h hdrfile] "
                    "tabfile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-p#      prediction field name             "
                    "(default: \"%s\")\n", res.col_pred);
    printf("-o#      significant digits for prediction "
                    "(default: %d)\n",     res.dig_pred);
    printf("-c#      confidence/probability field name "
                    "(default: no confidence)\n");
    printf("-z#      significant digits for confidence "
                    "(default: %d)\n",     res.dig_conf);
    printf("-t#      classification threshold          "
                    "(default: %g)\n",     thresh);
    printf("-a       align fields in output table      "
                    "(default: single separator)\n");
    printf("-w       do not write field names to output file\n");
    printf("-r#      record  separators                "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n"
           "         (only applicable with proper data table)\n");
    printf("regfile  file to read regression polynomial from\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write output table to (optional)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'p': optarg = &res.col_pred; break;
          case 'c': optarg = &res.col_conf; break;
          case 'o': res.dig_pred = (int)strtol(s, &s, 0); break;
          case 'z': res.dig_conf = (int)strtol(s, &s, 0); break;
          case 't': thresh = strtod(s, &s); break;
          case 'a': mout  |=  AS_ALIGN;     break;
          case 'w': mout  &= ~AS_ATT;       break;
          case 'r': optarg = &recseps;      break;
          case 'f': optarg = &fldseps;      break;
          case 'b': optarg = &blanks;       break;
          case 'C': optarg = &comment;      break;
          case 'n': mode  |= AS_WEIGHT;     break;
          case 'd': mode  |= AS_DFLT;       break;
          case 'h': optarg = &fn_hdr;       break;
          default : error(E_OPTION, *--s);  break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_reg = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument and */
  if ((k != 2) && (k != 3))     /* the number of arguments */
    error(E_ARGCNT);            /* (output file is optional) */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" for consistency */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_reg || !*fn_reg) i++;
  if  (!fn_tab || !*fn_tab) i++;
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if ((mout & AS_ATT) && (mout & AS_ALIGN))
    mout |= AS_ALNHDR;          /* set align to header flag */
  if (fn_out) mout |= AS_MARKED|AS_INFO1|AS_RDORD;
  else        mout  = 0;        /* set up the table write mode */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read regression polynomial --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the regression model */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_reg) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  matinp = (scn_first(scan) == T_ID)
        && (strcmp(scn_value(scan), "dom") != 0);
  if (matinp) m = -1;           /* if matrix version */
  else {                        /* if table version */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);/* create an attribute set */
    if (as_parse(attset, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan));
    a = as_attcnt(attset);      /* parse the domain descriptions */
    if (a <= 0) error(E_ATTCNT);/* get and check number of attributes */
    c = as_target(attset, NULL, -1);
    if (c >= a) c = a-1;        /* get the index */
    if (c <  0) c = 0;          /* of the target attribute and */
    res.att = as_att(attset, c);/* get it for result reporting */
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map and */
    am_target(attmap, c);       /* set the target attribute */
    m = am_inpcnt(attmap);      /* and get the number of inputs */
  }
  reg = reg_parse(scan, m+1,1); /* parse the regression polynomial */
  if (!reg || !scn_eof(scan,1)) /* check for end of file */
    error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* print a log message */
  fprintf(stderr, "[%"DIMID_FMT" parameter(s)]", reg_cnt(reg));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  if (matinp) {                 /* if input is a numerical matrix */
    /* --- process input matrix --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    if (fn_out) {               /* if to write an output file */
      twrite = twr_create();    /* create a table writer and */
      if (!twrite) error(E_NOMEM);   /* configure characters */
      twr_xchars(twrite, recseps, fldseps, blanks, "");
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
    }                           /* open the output file */
    k = vec_readx(&vec, &dim, tread);  /* read the first pattern */
    if (k) error(k, TRD_INFO(tread));  /* from the input file */
    m = reg_cnt(reg)-1;         /* get the number of inputs */
    if ((dim != m) && (dim != m+1))
      error(E_VECSIZE, dim);    /* check the pattern size */
    for (n = 0; k == 0; n++) {  /* pattern read loop */
      w = reg_exec(reg, vec);   /* execute regression polynomial */
      if (dim > m) {            /* and sum the squared errors */
        u = w -vec[m]; sse += u*u; }
      if (twrite) {             /* if to write an output table */
        for (c = 0; c < dim; c++) {
          twr_printf(twrite, "%-.*g", res.dig_conf, vec[c]);
          twr_fldsep(twrite);   /* print the pattern elements */
        }                       /* followed by a field separator */
        twr_printf(twrite, "%-.*g", res.dig_pred, w);
        twr_recsep(twrite);     /* print the prediction value */
      }                         /* and terminate the record */
      k = vec_read(vec, dim, tread);
    }                           /* read the next pattern */
    if (k < 0) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    if (twrite && (twr_delete(twrite, 1) < 0))
      error(E_FWRITE, fn_out);  /* delete the table writer */
    twrite = NULL;              /* and print a success message */
    fprintf(stderr, "[%"DIMID_FMT" vector(s)]", n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

    /* --- print error statistics --- */
    if (dim > m) {              /* if desired output is present */
      fprintf(stderr, "sse: %g, ", sse);
      if (n > 0) sse /= (double)n;
      fprintf(stderr, "mse: %g, rmse: %g\n", sse, sqrt(sse));
    } }                         /* print some error measures */

  else {                        /* if input is a data table */
    /* --- get target attribute --- */
    res.type = att_type(res.att);  /* get target attribute type */
    if (res.type != AT_NOM)     /* no confidence column */
      res.col_conf = NULL;      /* for numeric targets */
    as_setmark(attset, 1);      /* mark all attributes */
    att_setmark(res.att, 0);    /* except the class attribute */
    m = reg_cnt(reg)-1;         /* get the number of inputs */

    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      a = as_attcnt(attset);    /* get the number of attributes */
      fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s)]", a-1);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process table body --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    if (mout & AS_ALIGN) {      /* if to align the output columns */
      table = tab_create("table", attset, tpl_delete);
      if (!table) error(E_NOMEM);  /* read the data table */
      k = tab_read(table, tread, mode);
      if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
      trd_delete(tread, 1);     /* delete the table reader */
      tread = NULL;             /* and clear the variable */
      a = tab_attcnt(table);    /* get the number of attributes */
      n = tab_tplcnt(table);    /* and the number of tuples */
      w = tab_tplwgt(table);    /* and print a success message */
      fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s),", a-1);
      fprintf(stderr, " %"TPLID_FMT, n);
      if (w != (double)n) fprintf(stderr, "/%g", w);
      fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
      twrite = twr_create();       /* create a table writer and */
      if (!twrite) error(E_NOMEM); /* configure the characters */
      twr_xchars(twrite, recseps, fldseps, blanks, "");
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      fprintf(stderr, "writing %s ... ", twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      mout = AS_INST | (mout & ~AS_ATT);
      m  += (res.col_conf ? 2 : 1); /* add the output column(s) */
      vec = (double*)malloc((size_t)m *sizeof(double));
      if (!vec) error(E_NOMEM); /* create a data point/vector buffer */
      for (i = 0; i < n; i++) { /* traverse the tuples */
        tpl_toas(tab_tpl(table, i));
        am_exec(attmap, AM_BOTH, vec);
        predict(reg,vec,thresh);/* compute prediction for target */
        u = as_getwgt(attset);  /* get the tuple weight and */
        sse += res.err *u;      /* sum the prediction errors */
        if (as_write(attset, twrite, mout, infout) != 0) {
          free(vec); error(E_FWRITE, twr_name(twrite)); }
      }                         /* write the current tuple */
      free(vec); }              /* delete the vector buffer */
    else {                      /* if to process tuples directly */
      k = as_read(attset, tread, mode); /* read/generate table header */
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      if (!fn_out && (att_getmark(res.att) < 0))
        error(E_OUTPUT);        /* check for output to produce */
      if (fn_out) {             /* if to write an output file */
        twrite = twr_create();  /* create a table writer and */
        if (!twrite) error(E_NOMEM); /* configure the characters */
        twr_xchars(twrite, recseps, fldseps, blanks, "");
        t = clock();            /* start timer, open output file */
        if (twr_open(twrite, NULL, fn_out) != 0)
          error(E_FOPEN, twr_name(twrite));
        if ((mout & AS_ATT)     /* write a table header */
        &&  (as_write(attset, twrite, mout, infout) != 0))
          error(E_FWRITE, twr_name(twrite));
        mout = AS_INST | (mout & ~AS_ATT);
      }                         /* remove the attribute flag */
      vec = (double*)malloc((size_t)(m+1) *sizeof(double));
      if (!vec) error(E_NOMEM); /* create a data point/vector buffer */
      i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
      if (!(i & AS_DFLT))       /* if not done yet, read first tuple */
        k = as_read(attset, tread, mode);
      for (w = 0, n = 0; k == 0; n++) {
        am_exec(attmap, AM_BOTH, vec);
        predict(reg, vec, thresh);    /* classify the current tuple */
        w   += u = as_getwgt(attset); /* sum the tuple weights and */
        sse += res.err *u;      /* count the classification errors */
        if (twrite              /* write the current tuple */
        && (as_write(attset, twrite, mout, infout) != 0)) {
          free(vec); error(E_FWRITE, twr_name(twrite)); }
        k = as_read(attset, tread, mode);
      }                         /* try to read the next tuple */
      if (k < 0)                /* check for and report read error */
        error(-k, as_errmsg(attset, NULL, 0));
      trd_delete(tread, 1);     /* delete the table reader */
      tread = NULL;             /* and clear the variable */
      free(vec);                /* delete the vector buffer */
    }
    if (twrite) {               /* if an output file was written */
      if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
      twr_delete(twrite, 1);    /* close the output file and */
      twrite = NULL;            /* delete the table writer */
    }                           /* print a success message */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

    /* --- print error statistics --- */
    if (att_getmark(res.att) >= 0) {
      if (res.type != AT_NOM) { /* if the target att. is metric */
        fprintf(stderr, "sse: %g, ", sse);
        if (w > 0) sse /= w;    /* print some error measures */
        fprintf(stderr, "mse: %g, rmse: %g", sse, sqrt(sse)); }
      else {                    /* if the target att. is nominal */
        fprintf(stderr, "%g error(s) ", sse);
        fprintf(stderr, "(%.2f%%)", (w > 0) ? 100*(sse/w) : 100);
      }                         /* print an error indicator */
      fputc('\n', stderr);      /* terminate the error statistics */
    }
  }                             /* if (matinp) .. else .. */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
