/*----------------------------------------------------------------------
  File    : regress.h
  Contents: multivariate polynomial regression management
  Author  : Christian Borgelt
  History : 2001.10.13 file created
            2001.10.15 first version completed
            2001.10.19 function/macro reg_coeff added
            2001.10.20 function reg_solve redesigned
            2001.10.26 function reg_sse added, reg_solve modified
            2007.04.04 extended regression functions added
            2007.04.12 functions reg_parse and reg_parsex added
            2013.05.06 Tikhonov regularization parameter added
            2013.08.13 adapted to definition of type DIMID in matrix.h
            2015.07.30 logistic regression with maximum likelihood added
            2015.07.31 functions reg_minmax() and reg_reglim() added
            2018.09.13 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#ifndef __REGRESS__
#define __REGRESS__
#include <stdio.h>
#include "matrix.h"
#ifdef REG_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#include "scanner.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- creation modes --- */
#define REG_MARKED    AM_MARKED /* use only marked attributes */

/* --- optimization modes --- */
#define REG_NONE      0         /* no subsequent optimization */
#define REG_ERR_GD    1         /* gradient descent on error function */
#define REG_ERR_NR    2         /* Newton-Raphson   on error function */
#define REG_LLH_GA    3         /* gradient ascent  on likelihood */
#define REG_LLH_NR    4         /* Newton-Raphson   on likelihood */

/* --- description modes --- */
#define REG_TITLE     0x0001    /* print a title (as a comment) */
#define REG_EXPOS     0x0002    /* print exponents of variables */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a regression object --- */
  int    deg;                   /* degree of polynomial */
  DIMID  dim;                   /* number of dimensions */
  DIMID  cnt;                   /* number of regression coefficients */
  DIMID  *tab;                  /* table of section sizes */
  double *vec;                  /* buffer for an input vector */
  double *buf;                  /* buffer for the power products */
  MATRIX *mat;                  /* regression matrix */
  double *rhs;                  /* right hand side of equation system */
  double *cfs;                  /* regression coefficients */
  double so2;                   /* sum of squared output values */
  double min;                   /* minimum for logit transformation */
  double max;                   /* maximum for logit transformation */
  double reg;                   /* Tikhonov regularization parameter */
  double lim;                   /* limit for Tikhonov regularization */
  double sse;                   /* sum of squared errors */
} REGRESS;                      /* (regression object) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern REGRESS* reg_create  (DIMID dim, int deg);
extern void     reg_delete  (REGRESS *reg);

extern void     reg_setmin  (REGRESS *reg, double min);
extern double   reg_getmin  (REGRESS *reg);
extern void     reg_setmax  (REGRESS *reg, double max);
extern double   reg_getmax  (REGRESS *reg);
extern void     reg_minmax  (REGRESS *reg, double min, double max);
extern int      reg_logit   (REGRESS *reg);
extern void     reg_setreg  (REGRESS *reg, double val);
extern double   reg_getreg  (REGRESS *reg);
extern void     reg_setlim  (REGRESS *reg, double lim);
extern double   reg_getlim  (REGRESS *reg);
extern void     reg_reglim  (REGRESS *reg, double val, double lim);
extern void     reg_init    (REGRESS *reg);

extern int      reg_deg     (REGRESS *reg);
extern DIMID    reg_dim     (REGRESS *reg);
extern DIMID    reg_cnt     (REGRESS *reg);
extern double   reg_coeff   (REGRESS *reg, DIMID index);
extern double   reg_wgtsum  (REGRESS *reg);

extern int      reg_aggvec  (REGRESS *reg, double *vec, double wgt);
extern DIMID    reg_aggmat  (REGRESS *reg, MATRIX *data);
extern int      reg_solve   (REGRESS *reg);
extern DIMID    reg_logopt  (REGRESS *reg, MATRIX *data, int mode,
                             double lrate, double limit, int maxiter);
extern double   reg_exec    (REGRESS *reg, double *vec);
extern double   reg_sse     (REGRESS *reg);

extern int      reg_desc    (REGRESS *reg, FILE *file,
                             int mode, int maxlen);
#ifdef REG_PARSE
extern REGRESS* reg_parse   (SCANNER *scan, DIMID dim, int hdr);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define reg_setmin(r,x)     ((r)->min = (x))
#define reg_getmin(r)       ((r)->min)
#define reg_setmax(r,x)     ((r)->max = (x))
#define reg_getmax(r)       ((r)->max)
#define reg_logit(r)        ((r)->max > (r)->min)
#define reg_setreg(r,x)     ((r)->reg = (x))
#define reg_getreg(r)       ((r)->reg)
#define reg_setlim(r,x)     ((r)->lim = (x))
#define reg_getlim(r)       ((r)->lim)
#define reg_dim(r)          ((r)->dim)
#define reg_deg(r)          ((r)->deg)
#define reg_cnt(r)          ((r)->cnt)
#define reg_coeff(r,i)      ((r)->cfs[i])
#define reg_wgtsum(r)       mat_wgtsum((r)->mat)

#endif
